'''
Contains all the commonly used functions and data
useful for multiple tmd modules.
'''

term_dict = {'x': 0,
             'y': 1,
             'z': 2}

TREE_TYPE_DICT = {1: 'soma',
                  2: 'axon',
                  3: 'basal',
                  4: 'apical'}

SOMA_TYPE = 1
